@@ -1,4 +1,4 @@
-// Copyright 2019-2023 The NATS Authors
+// Copyright 2019-2024 The NATS Authors
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -22,6 +22,7 @@ import (
 	"time"
 
 	"github.com/nats-io/nats-server/v2/server/avl"
+	"github.com/nats-io/nats-server/v2/server/stree"
 )
 
 // TODO(dlc) - This is a fairly simplistic approach but should do for now.
@@ -30,7 +31,7 @@ type memStore struct {
 	cfg         StreamConfig
 	state       StreamState
 	msgs        map[uint64]*StoreMsg
-	fss         map[string]*SimpleState
+	fss         *stree.SubjectTree[SimpleState]
 	dmap        avl.SequenceSet
 	maxp        int64
 	scb         StorageUpdateHandler
@@ -48,7 +49,7 @@ func newMemStore(cfg *StreamConfig) (*memStore, error) {
 	}
 	ms := &memStore{
 		msgs: make(map[uint64]*StoreMsg),
-		fss:  make(map[string]*SimpleState),
+		fss:  stree.NewSubjectTree[SimpleState](),
 		maxp: cfg.MaxMsgsPer,
 		cfg:  *cfg,
 	}
@@ -88,11 +89,12 @@ func (ms *memStore) UpdateConfig(cfg *StreamConfig) error {
 	// If the value is smaller we need to enforce that.
 	if ms.maxp != 0 && ms.maxp < maxp {
 		lm := uint64(ms.maxp)
-		for subj, ss := range ms.fss {
+		ms.fss.Iter(func(subj []byte, ss *SimpleState) bool {
 			if ss.Msgs > lm {
-				ms.enforcePerSubjectLimit(subj, ss)
+				ms.enforcePerSubjectLimit(bytesToString(subj), ss)
 			}
-		}
+			return true
+		})
 	}
 	ms.mu.Unlock()
 
@@ -113,7 +115,8 @@ func (ms *memStore) storeRawMsg(subj string, hdr, msg []byte, seq uint64, ts int
 	var ss *SimpleState
 	var asl bool
 	if len(subj) > 0 {
-		if ss = ms.fss[subj]; ss != nil {
+		var ok bool
+		if ss, ok = ms.fss.Find(stringToBytes(subj)); ok {
 			asl = ms.maxp > 0 && ss.Msgs >= uint64(ms.maxp)
 		}
 	}
@@ -123,23 +126,27 @@ func (ms *memStore) storeRawMsg(subj string, hdr, msg []byte, seq uint64, ts int
 		if asl && ms.cfg.DiscardNewPer {
 			return ErrMaxMsgsPerSubject
 		}
-		if ms.cfg.MaxMsgs > 0 && ms.state.Msgs >= uint64(ms.cfg.MaxMsgs) {
-			// If we are tracking max messages per subject and are at the limit we will replace, so this is ok.
-			if !asl {
-				return ErrMaxMsgs
+		// If we are discard new and limits policy and clustered, we do the enforcement
+		// above and should not disqualify the message here since it could cause replicas to drift.
+		if ms.cfg.Retention == LimitsPolicy || ms.cfg.Replicas == 1 {
+			if ms.cfg.MaxMsgs > 0 && ms.state.Msgs >= uint64(ms.cfg.MaxMsgs) {
+				// If we are tracking max messages per subject and are at the limit we will replace, so this is ok.
+				if !asl {
+					return ErrMaxMsgs
+				}
 			}
-		}
-		if ms.cfg.MaxBytes > 0 && ms.state.Bytes+memStoreMsgSize(subj, hdr, msg) >= uint64(ms.cfg.MaxBytes) {
-			if !asl {
-				return ErrMaxBytes
-			}
-			// If we are here we are at a subject maximum, need to determine if dropping last message gives us enough room.
-			if ss.firstNeedsUpdate {
-				ms.recalculateFirstForSubj(subj, ss.First, ss)
-			}
-			sm, ok := ms.msgs[ss.First]
-			if !ok || memStoreMsgSize(sm.subj, sm.hdr, sm.msg) < memStoreMsgSize(subj, hdr, msg) {
-				return ErrMaxBytes
+			if ms.cfg.MaxBytes > 0 && ms.state.Bytes+memStoreMsgSize(subj, hdr, msg) >= uint64(ms.cfg.MaxBytes) {
+				if !asl {
+					return ErrMaxBytes
+				}
+				// If we are here we are at a subject maximum, need to determine if dropping last message gives us enough room.
+				if ss.firstNeedsUpdate {
+					ms.recalculateFirstForSubj(subj, ss.First, ss)
+				}
+				sm, ok := ms.msgs[ss.First]
+				if !ok || memStoreMsgSize(sm.subj, sm.hdr, sm.msg) < memStoreMsgSize(subj, hdr, msg) {
+					return ErrMaxBytes
+				}
 			}
 		}
 	}
@@ -191,7 +198,7 @@ func (ms *memStore) storeRawMsg(subj string, hdr, msg []byte, seq uint64, ts int
 				ms.enforcePerSubjectLimit(subj, ss)
 			}
 		} else {
-			ms.fss[subj] = &SimpleState{Msgs: 1, First: seq, Last: seq}
+			ms.fss.Insert([]byte(subj), SimpleState{Msgs: 1, First: seq, Last: seq})
 		}
 	}
 
@@ -256,12 +263,42 @@ func (ms *memStore) SkipMsg() uint64 {
 	if ms.state.Msgs == 0 {
 		ms.state.FirstSeq = seq
 		ms.state.FirstTime = now
+	} else {
+		ms.dmap.Insert(seq)
 	}
-	ms.updateFirstSeq(seq)
 	ms.mu.Unlock()
 	return seq
 }
 
+// Skip multiple msgs.
+func (ms *memStore) SkipMsgs(seq uint64, num uint64) error {
+	// Grab time.
+	now := time.Now().UTC()
+
+	ms.mu.Lock()
+	defer ms.mu.Unlock()
+
+	// Check sequence matches our last sequence.
+	if seq != ms.state.LastSeq+1 {
+		if seq > 0 {
+			return ErrSequenceMismatch
+		}
+		seq = ms.state.LastSeq + 1
+	}
+	lseq := seq + num - 1
+
+	ms.state.LastSeq = lseq
+	ms.state.LastTime = now
+	if ms.state.Msgs == 0 {
+		ms.state.FirstSeq, ms.state.FirstTime = lseq+1, now
+	} else {
+		for ; seq <= lseq; seq++ {
+			ms.dmap.Insert(seq)
+		}
+	}
+	return nil
+}
+
 // RegisterStorageUpdates registers a callback for updates to storage changes.
 // It will present number of messages and bytes as a signed integer and an
 // optional sequence number of the message if a single.
@@ -333,14 +370,17 @@ func (ms *memStore) filteredStateLocked(sseq uint64, filter string, lastPerSubje
 		return ss
 	}
 
-	isAll := filter == _EMPTY_ || filter == fwcs
+	if filter == _EMPTY_ {
+		filter = fwcs
+	}
+	isAll := filter == fwcs
 
 	// First check if we can optimize this part.
 	// This means we want all and the starting sequence was before this block.
 	if isAll && sseq <= ms.state.FirstSeq {
 		total := ms.state.Msgs
 		if lastPerSubject {
-			total = uint64(len(ms.fss))
+			total = uint64(ms.fss.Size())
 		}
 		return SimpleState{
 			Msgs:  total,
@@ -356,7 +396,7 @@ func (ms *memStore) filteredStateLocked(sseq uint64, filter string, lastPerSubje
 
 	// 1. See if we match any subs from fss.
 	// 2. If we match and the sseq is past ss.Last then we can use meta only.
-	// 3. If we match and we need to do a partial, break and clear any totals and do a full scan like num pending.
+	// 3. If we match we need to do a partial, break and clear any totals and do a full scan like num pending.
 
 	isMatch := func(subj string) bool {
 		if isAll {
@@ -385,21 +425,20 @@ func (ms *memStore) filteredStateLocked(sseq uint64, filter string, lastPerSubje
 
 	var havePartial bool
 	// We will track start and end sequences as we go.
-	for subj, fss := range ms.fss {
-		if isMatch(subj) {
-			if fss.firstNeedsUpdate {
-				ms.recalculateFirstForSubj(subj, fss.First, fss)
-			}
-			if sseq <= fss.First {
-				update(fss)
-			} else if sseq <= fss.Last {
-				// We matched but its a partial.
-				havePartial = true
-				// Don't break here, we will update to keep tracking last.
-				update(fss)
-			}
+	ms.fss.Match(stringToBytes(filter), func(subj []byte, fss *SimpleState) {
+		subjs := bytesToString(subj)
+		if fss.firstNeedsUpdate {
+			ms.recalculateFirstForSubj(subjs, fss.First, fss)
 		}
-	}
+		if sseq <= fss.First {
+			update(fss)
+		} else if sseq <= fss.Last {
+			// We matched but it is a partial.
+			havePartial = true
+			// Don't break here, we will update to keep tracking last.
+			update(fss)
+		}
+	})
 
 	// If we did not encounter any partials we can return here.
 	if !havePartial {
@@ -409,7 +448,7 @@ func (ms *memStore) filteredStateLocked(sseq uint64, filter string, lastPerSubje
 	// If we are here we need to scan the msgs.
 	// Capture first and last sequences for scan and then clear what we had.
 	first, last := ss.First, ss.Last
-	// To track if we decide to exclude and we need to calculate first.
+	// To track if we decide to exclude we need to calculate first.
 	var needScanFirst bool
 	if first < sseq {
 		first = sseq
@@ -446,7 +485,7 @@ func (ms *memStore) filteredStateLocked(sseq uint64, filter string, lastPerSubje
 		for seq := ms.state.FirstSeq; seq < first; seq++ {
 			if sm, ok := ms.msgs[seq]; ok && !seen[sm.subj] && isMatch(sm.subj) {
 				if lastPerSubject {
-					tss = ms.fss[sm.subj]
+					tss, _ = ms.fss.Find(stringToBytes(sm.subj))
 				}
 				// If we are last per subject, make sure to only adjust if all messages are before our first.
 				if tss == nil || tss.Last < first {
@@ -485,26 +524,29 @@ func (ms *memStore) SubjectsState(subject string) map[string]SimpleState {
 	ms.mu.RLock()
 	defer ms.mu.RUnlock()
 
-	if len(ms.fss) == 0 {
+	if ms.fss.Size() == 0 {
 		return nil
 	}
 
-	fss := make(map[string]SimpleState)
-	for subj, ss := range ms.fss {
-		if subject == _EMPTY_ || subject == fwcs || subjectIsSubsetMatch(subj, subject) {
-			if ss.firstNeedsUpdate {
-				ms.recalculateFirstForSubj(subj, ss.First, ss)
-			}
-			oss := fss[subj]
-			if oss.First == 0 { // New
-				fss[subj] = *ss
-			} else {
-				// Merge here.
-				oss.Last, oss.Msgs = ss.Last, oss.Msgs+ss.Msgs
-				fss[subj] = oss
-			}
-		}
+	if subject == _EMPTY_ {
+		subject = fwcs
 	}
+
+	fss := make(map[string]SimpleState)
+	ms.fss.Match(stringToBytes(subject), func(subj []byte, ss *SimpleState) {
+		subjs := string(subj)
+		if ss.firstNeedsUpdate {
+			ms.recalculateFirstForSubj(subjs, ss.First, ss)
+		}
+		oss := fss[subjs]
+		if oss.First == 0 { // New
+			fss[subjs] = *ss
+		} else {
+			// Merge here.
+			oss.Last, oss.Msgs = ss.Last, oss.Msgs+ss.Msgs
+			fss[subjs] = oss
+		}
+	})
 	return fss
 }
 
@@ -513,7 +555,7 @@ func (ms *memStore) SubjectsTotals(filterSubject string) map[string]uint64 {
 	ms.mu.RLock()
 	defer ms.mu.RUnlock()
 
-	if len(ms.fss) == 0 {
+	if ms.fss.Size() == 0 {
 		return nil
 	}
 
@@ -523,15 +565,16 @@ func (ms *memStore) SubjectsTotals(filterSubject string) map[string]uint64 {
 	isAll := filterSubject == _EMPTY_ || filterSubject == fwcs
 
 	fst := make(map[string]uint64)
-	for subj, ss := range ms.fss {
+	ms.fss.Match(stringToBytes(filterSubject), func(subj []byte, ss *SimpleState) {
+		subjs := string(subj)
 		if isAll {
-			fst[subj] = ss.Msgs
+			fst[subjs] = ss.Msgs
 		} else {
-			if tts := tokenizeSubjectIntoSlice(tsa[:0], subj); isSubsetMatchTokenized(tts, fts) {
-				fst[subj] = ss.Msgs
+			if tts := tokenizeSubjectIntoSlice(tsa[:0], subjs); isSubsetMatchTokenized(tts, fts) {
+				fst[subjs] = ss.Msgs
 			}
 		}
-	}
+	})
 	return fst
 }
 
@@ -605,7 +648,11 @@ func (ms *memStore) resetAgeChk(delta int64) {
 
 	fireIn := ms.cfg.MaxAge
 	if delta > 0 && time.Duration(delta) < fireIn {
-		fireIn = time.Duration(delta)
+		if fireIn = time.Duration(delta); fireIn < 250*time.Millisecond {
+			// Only fire at most once every 250ms.
+			// Excessive firing can effect ingest performance.
+			fireIn = time.Second
+		}
 	}
 	if ms.ageChk != nil {
 		ms.ageChk.Reset(fireIn)
@@ -616,19 +663,21 @@ func (ms *memStore) resetAgeChk(delta int64) {
 
 // Will expire msgs that are too old.
 func (ms *memStore) expireMsgs() {
-	ms.mu.Lock()
-	defer ms.mu.Unlock()
-
+	ms.mu.RLock()
 	now := time.Now().UnixNano()
 	minAge := now - int64(ms.cfg.MaxAge)
+	ms.mu.RUnlock()
+
 	for {
+		ms.mu.Lock()
 		if sm, ok := ms.msgs[ms.state.FirstSeq]; ok && sm.ts <= minAge {
 			ms.deleteFirstMsgOrPanic()
 			// Recalculate in case we are expiring a bunch.
 			now = time.Now().UnixNano()
 			minAge = now - int64(ms.cfg.MaxAge)
-
+			ms.mu.Unlock()
 		} else {
+			// We will exit here
 			if len(ms.msgs) == 0 {
 				if ms.ageChk != nil {
 					ms.ageChk.Stop()
@@ -647,7 +696,8 @@ func (ms *memStore) expireMsgs() {
 					ms.ageChk = time.AfterFunc(fireIn, ms.expireMsgs)
 				}
 			}
-			return
+			ms.mu.Unlock()
+			break
 		}
 	}
 }
@@ -725,7 +775,7 @@ func (ms *memStore) purge(fseq uint64) (uint64, error) {
 	ms.state.Bytes = 0
 	ms.state.Msgs = 0
 	ms.msgs = make(map[uint64]*StoreMsg)
-	ms.fss = make(map[string]*SimpleState)
+	ms.fss = stree.NewSubjectTree[SimpleState]()
 	ms.mu.Unlock()
 
 	if cb != nil {
@@ -748,15 +798,16 @@ func (ms *memStore) Compact(seq uint64) (uint64, error) {
 	ms.mu.Lock()
 	cb := ms.scb
 	if seq <= ms.state.LastSeq {
-		sm, ok := ms.msgs[seq]
-		if !ok {
-			ms.mu.Unlock()
-			return 0, ErrStoreMsgNotFound
+		fseq := ms.state.FirstSeq
+		// Determine new first sequence.
+		for ; seq <= ms.state.LastSeq; seq++ {
+			if sm, ok := ms.msgs[seq]; ok {
+				ms.state.FirstSeq = seq
+				ms.state.FirstTime = time.Unix(0, sm.ts).UTC()
+				break
+			}
 		}
-		ms.state.FirstSeq = seq
-		ms.state.FirstTime = time.Unix(0, sm.ts).UTC()
-
-		for seq := seq - 1; seq > 0; seq-- {
+		for seq := seq - 1; seq >= fseq; seq-- {
 			if sm := ms.msgs[seq]; sm != nil {
 				bytes += memStoreMsgSize(sm.subj, sm.hdr, sm.msg)
 				purged++
@@ -816,7 +867,7 @@ func (ms *memStore) reset() error {
 	ms.state.Bytes = 0
 	// Reset msgs and fss.
 	ms.msgs = make(map[uint64]*StoreMsg)
-	ms.fss = make(map[string]*SimpleState)
+	ms.fss = stree.NewSubjectTree[SimpleState]()
 
 	ms.mu.Unlock()
 
@@ -920,7 +971,8 @@ func (ms *memStore) LoadLastMsg(subject string, smp *StoreMsg) (*StoreMsg, error
 	if subject == _EMPTY_ || subject == fwcs {
 		sm, ok = ms.msgs[ms.state.LastSeq]
 	} else if subjectIsLiteral(subject) {
-		if ss := ms.fss[subject]; ss != nil && ss.Msgs > 0 {
+		var ss *SimpleState
+		if ss, ok = ms.fss.Find(stringToBytes(subject)); ok && ss.Msgs > 0 {
 			sm, ok = ms.msgs[ss.Last]
 		}
 	} else if ss := ms.filteredStateLocked(1, subject, true); ss.Msgs > 0 {
@@ -937,6 +989,40 @@ func (ms *memStore) LoadLastMsg(subject string, smp *StoreMsg) (*StoreMsg, error
 	return smp, nil
 }
 
+// LoadNextMsgMulti will find the next message matching any entry in the sublist.
+func (ms *memStore) LoadNextMsgMulti(sl *Sublist, start uint64, smp *StoreMsg) (sm *StoreMsg, skip uint64, err error) {
+	// TODO(dlc) - for now simple linear walk to get started.
+	ms.mu.RLock()
+	defer ms.mu.RUnlock()
+
+	if start < ms.state.FirstSeq {
+		start = ms.state.FirstSeq
+	}
+
+	// If past the end no results.
+	if start > ms.state.LastSeq || ms.state.Msgs == 0 {
+		return nil, ms.state.LastSeq, ErrStoreEOF
+	}
+
+	// Initial setup.
+	fseq, lseq := start, ms.state.LastSeq
+
+	for nseq := fseq; nseq <= lseq; nseq++ {
+		sm, ok := ms.msgs[nseq]
+		if !ok {
+			continue
+		}
+		if sl.HasInterest(sm.subj) {
+			if smp == nil {
+				smp = new(StoreMsg)
+			}
+			sm.copy(smp)
+			return smp, nseq, nil
+		}
+	}
+	return nil, ms.state.LastSeq, ErrStoreEOF
+}
+
 // LoadNextMsg will find the next message matching the filter subject starting at the start sequence.
 // The filter subject can be a wildcard.
 func (ms *memStore) LoadNextMsg(filter string, wc bool, start uint64, smp *StoreMsg) (*StoreMsg, uint64, error) {
@@ -948,16 +1034,19 @@ func (ms *memStore) LoadNextMsg(filter string, wc bool, start uint64, smp *Store
 	}
 
 	// If past the end no results.
-	if start > ms.state.LastSeq {
+	if start > ms.state.LastSeq || ms.state.Msgs == 0 {
 		return nil, ms.state.LastSeq, ErrStoreEOF
 	}
 
-	isAll := filter == _EMPTY_ || filter == fwcs
+	if filter == _EMPTY_ {
+		filter = fwcs
+	}
+	isAll := filter == fwcs
 
-	// Skip scan of ms.fss is number of messages in the block are less than
+	// Skip scan of ms.fss if number of messages in the block are less than
 	// 1/2 the number of subjects in ms.fss. Or we have a wc and lots of fss entries.
 	const linearScanMaxFSS = 256
-	doLinearScan := isAll || 2*int(ms.state.LastSeq-start) < len(ms.fss) || (wc && len(ms.fss) > linearScanMaxFSS)
+	doLinearScan := isAll || 2*int(ms.state.LastSeq-start) < ms.fss.Size() || (wc && ms.fss.Size() > linearScanMaxFSS)
 
 	// Initial setup.
 	fseq, lseq := start, ms.state.LastSeq
@@ -966,16 +1055,14 @@ func (ms *memStore) LoadNextMsg(filter string, wc bool, start uint64, smp *Store
 		subs := []string{filter}
 		if wc || isAll {
 			subs = subs[:0]
-			for fsubj := range ms.fss {
-				if isAll || subjectIsSubsetMatch(fsubj, filter) {
-					subs = append(subs, fsubj)
-				}
-			}
+			ms.fss.Match(stringToBytes(filter), func(subj []byte, val *SimpleState) {
+				subs = append(subs, string(subj))
+			})
 		}
 		fseq, lseq = ms.state.LastSeq, uint64(0)
 		for _, subj := range subs {
-			ss := ms.fss[subj]
-			if ss == nil {
+			ss, ok := ms.fss.Find(stringToBytes(subj))
+			if !ok {
 				continue
 			}
 			if ss.firstNeedsUpdate {
@@ -1058,20 +1145,17 @@ func (ms *memStore) updateFirstSeq(seq uint64) {
 			ms.dmap.Delete(seq)
 		}
 	}
-	if ms.dmap.IsEmpty() {
-		ms.dmap.SetInitialMin(ms.state.FirstSeq)
-	}
 }
 
 // Remove a seq from the fss and select new first.
 // Lock should be held.
 func (ms *memStore) removeSeqPerSubject(subj string, seq uint64) {
-	ss := ms.fss[subj]
-	if ss == nil {
+	ss, ok := ms.fss.Find(stringToBytes(subj))
+	if !ok {
 		return
 	}
 	if ss.Msgs == 1 {
-		delete(ms.fss, subj)
+		ms.fss.Delete(stringToBytes(subj))
 		return
 	}
 	ss.Msgs--
@@ -1089,7 +1173,7 @@ func (ms *memStore) removeSeqPerSubject(subj string, seq uint64) {
 	}
 }
 
-// Will recalulate the first sequence for this subject in this block.
+// Will recalculate the first sequence for this subject in this block.
 // Lock should be held.
 func (ms *memStore) recalculateFirstForSubj(subj string, startSeq uint64, ss *SimpleState) {
 	for tseq := startSeq + 1; tseq <= ss.Last; tseq++ {
@@ -1102,7 +1186,7 @@ func (ms *memStore) recalculateFirstForSubj(subj string, startSeq uint64, ss *Si
 }
 
 // Removes the message referenced by seq.
-// Lock should he held.
+// Lock should be held.
 func (ms *memStore) removeMsg(seq uint64, secure bool) bool {
 	var ss uint64
 	sm, ok := ms.msgs[seq]
@@ -1171,7 +1255,7 @@ func (ms *memStore) FastState(state *StreamState) {
 		}
 	}
 	state.Consumers = ms.consumers
-	state.NumSubjects = len(ms.fss)
+	state.NumSubjects = ms.fss.Size()
 	ms.mu.RUnlock()
 }
 
@@ -1181,18 +1265,21 @@ func (ms *memStore) State() StreamState {
 
 	state := ms.state
 	state.Consumers = ms.consumers
-	state.NumSubjects = len(ms.fss)
+	state.NumSubjects = ms.fss.Size()
 	state.Deleted = nil
 
 	// Calculate interior delete details.
 	if numDeleted := int((state.LastSeq - state.FirstSeq + 1) - state.Msgs); numDeleted > 0 {
-		state.Deleted = make([]uint64, 0, state.NumDeleted)
-		// TODO(dlc) - Too Simplistic, once state is updated to allow runs etc redo.
-		for seq := state.FirstSeq + 1; seq < ms.state.LastSeq; seq++ {
-			if _, ok := ms.msgs[seq]; !ok {
+		state.Deleted = make([]uint64, 0, numDeleted)
+		fseq, lseq := state.FirstSeq, state.LastSeq
+		ms.dmap.Range(func(seq uint64) bool {
+			if seq < fseq || seq > lseq {
+				ms.dmap.Delete(seq)
+			} else {
 				state.Deleted = append(state.Deleted, seq)
 			}
-		}
+			return true
+		})
 	}
 	if len(state.Deleted) > 0 {
 		state.NumDeleted = len(state.Deleted)
@@ -1315,19 +1402,25 @@ func (ms *memStore) EncodedStreamState(failed uint64) ([]byte, error) {
 
 // SyncDeleted will make sure this stream has same deleted state as dbs.
 func (ms *memStore) SyncDeleted(dbs DeleteBlocks) {
+	ms.mu.Lock()
+	defer ms.mu.Unlock()
+
 	// For now we share one dmap, so if we have one entry here check if states are the same.
 	// Note this will work for any DeleteBlock type, but we expect this to be a dmap too.
 	if len(dbs) == 1 {
-		ms.mu.RLock()
 		min, max, num := ms.dmap.State()
-		ms.mu.RUnlock()
 		if pmin, pmax, pnum := dbs[0].State(); pmin == min && pmax == max && pnum == num {
 			return
 		}
 	}
+	lseq := ms.state.LastSeq
 	for _, db := range dbs {
-		db.Range(func(dseq uint64) bool {
-			ms.RemoveMsg(dseq)
+		// Skip if beyond our current state.
+		if first, _, _ := db.State(); first > lseq {
+			continue
+		}
+		db.Range(func(seq uint64) bool {
+			ms.removeMsg(seq, false)
 			return true
 		})
 	}
@@ -1438,7 +1531,7 @@ func (o *consumerMemStore) UpdateDelivered(dseq, sseq, dc uint64, ts int64) erro
 			if o.state.Redelivered == nil {
 				o.state.Redelivered = make(map[uint64]uint64)
 			}
-			// Only update if greater then what we already have.
+			// Only update if greater than what we already have.
 			if o.state.Redelivered[sseq] < dc-1 {
 				o.state.Redelivered[sseq] = dc - 1
 			}
@@ -1589,7 +1682,7 @@ func (o *consumerMemStore) stateWithCopy(doCopy bool) (*ConsumerState, error) {
 	return state, nil
 }
 
-// EncodeState for this consumer store.
+// EncodedState for this consumer store.
 func (o *consumerMemStore) EncodedState() ([]byte, error) {
 	o.mu.Lock()
 	defer o.mu.Unlock()
